/*
 * WiPal - A library and a set of tools to manipulate wireless traces.
 * Copyright (C) 2007, 2008  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */
#ifndef PCAP_DESCRIPTOR_HH_
# define PCAP_DESCRIPTOR_HH_

extern "C"
{
# include <inttypes.h>
}
# include <vector>
# include <ios>
# include <string>
# include <cstdlib>

# include <wipal/pcap/frame_iterator.hh>
# include <wipal/tool/iterable.hh>
# include <wipal/pcap/dumper.hh>

/**
 * PCAP files input/output.
 *
 * This namespace holds the compounds required to perform PCAP (Packet
 * CAPture) files input/output.
 *
 * - In order to <b>read PCAP files</b>, one needs a @link descriptor
 *   PCAP descriptor @endlink object.  PCAP descriptors also enable
 *   users to instanciate dumper objects.
 * - In order to <b>write PCAP files</b>, one needs a dumper object.
 * - It is possible to <b>consider separate PCAP files as one unique
 *   trace</b> consisting of the concatenation of all files.  Do this
 *   with a list_of_traces object.
 * - Use a match object to <b>look for specific byte sequences</b>
 *   into @link frame_descriptor frame descriptors @endlink.
 *
 * @see descriptor, dumper, list_of_traces, match, frame_descriptor.
 */
namespace pcapxx
{

  template <class D>
  struct descriptor;

} // End of namespace pcapxx.

namespace tool
{

  /// Virtual %types for pcapxx::descriptor.  @see types
  template <class B>
  struct types< pcapxx::descriptor<B> >
  {
    /// @link pcapxx::descriptor PCAP descriptor @endlink exact type.
    typedef WP_GET_EXACT(B, pcapxx::descriptor<B>)		exact_type;

    /// @link pcapxx::descriptor PCAP descriptor @endlink iterator type.
    typedef pcapxx::internals::frame_iterator<exact_type>	iterator;
  };

} // End of namespace tool.

namespace pcapxx
{

  /**
   * Enable various PCAP operations.
   *
   * A PCAP %descriptor allows you to read frames (or packets) from
   * PCAP files.  You may also use a PCAP %descriptor to retrieve
   * a dumper object, in order to dump frames (or packets) into a PCAP
   * file.
   *
   * A PCAP %descriptor is an iterable object.  You may iterate over
   * frames like this:
   *
   * @code
   * pcapxx::descriptor<> d ("file.pcap");
   *
   * for (pcapxx::descriptor<>::iterator i = d.begin(); i != d.end(); ++i)
   *   {
   *     // Access to i->pcap_header and i->bytes.
   *     //
   *     // *i is of type frame_descriptor.
   *   }
   * @endcode
   *
   * @see dumper, frame_descriptor, internals::frame_iterator
   */
  template <class Bottom = tool::bottom>
  struct descriptor: WP_INHERIT(public tool::iterable, descriptor<Bottom>)
  {
    /// @name Static hierarchies related types.
    //@{
    typedef WP_GET_EXACT(Bottom, descriptor<Bottom>)	exact_type;
    typedef tool::iterable<exact_type>			super_type;
    //@}

    /// Iterator type for this iterable class.
    typedef WP_TYPE(iterator, exact_type)		iterator;
    typedef iterator					const_iterator;

    /**
     * Internal index type for random frame access.
     *
     * Only advanced users need to know about this.
     *
     * @see mark_step.
     */
    typedef std::vector<std::streampos>			index_type;

    enum
    {
      /**
       * Register a mark every @c mark_step packets.
       *
       * A mark is an offset in the trace file.  Marks are
       * stored in a vector at regular intervals, so fast
       * access to random packets is possible.
       *
       * This is a trade-off: more marks allows faster access, but
       * fewer marks uses less memory.
       */
      mark_step	= 4096,

      /**
       * When loading traces, read packet data with steps of @c
       * junk_len bytes.
       *
       * The more the better, but one must be careful not to grow
       * the stack too much.  We use reads here instead of seeks
       * not to break the underlying C++ buffering.  Tests showed
       * this is faster.
       */
      junk_len	= 4096
    };

    /// PCAP file link types.
    enum link_type
    {
      LOOP			= 0,	///< BSD loopback encapsulation
      EN10MB			= 1,	///< Ethernet (10Mb)
      EN3MB			= 2,	///< Experimental Ethernet (3Mb)
      AX25			= 3,	///< Amateur Radio AX.25
      PRONET			= 4,	///< Proteon ProNET Token Ring
      CHAOS			= 5,	///< Chaos
      IEEE802			= 6,	///< IEEE 802 Networks
      ARCNET			= 7,	///< ARCNET, with BSD-style header
      SLIP			= 8,	///< Serial Line IP
      PPP			= 9,	///< Point-to-point Protocol
      FDDI			= 10,	///< FDDI
      IEEE802_11		= 105,	///< IEEE 802.11 wireless
      PRISM_HEADER		= 119,	///< Prism PHY-layer information
      IEEE802_11_RADIO		= 127,	///< Radiotap header
      IEEE802_11_RADIO_AVS	= 163	///< AVS header
    };

    /// Construct a %descriptor from a file name.
    descriptor(const std::string& filename, bool build_index = false);

    /**
     * Construct a %descriptor from a file name; use an already
     * constructed index.
     */
    descriptor(const std::string&	filename,
	       const index_type&	marks,
	       unsigned			pkt_count);

    /**
     * Construct a %descriptor from a file name, with progress
     * indicator callbacks.
     *
     * @param[in]	filename	The file's name.
     * @param[in]	pi_setup	Called before reading packets with
     *					two arguments.  The first one is
     *					the first packet's offset, the last
     *					one is the file size.
     * @param[in]	pi_update	Called periodically when reading
     *					packets.  The current offset in the
     *					file is passed as argument.  Returns
     *					either 0 (continue) or a pointer to
     *					a string (stop and throws
     *					@c interrupted(string) ).
     */
    template <typename ProgressInfoSetupType,
	      typename ProgressInfoUpdateType>
    descriptor(const std::string&	filename,
	       ProgressInfoSetupType&	pi_setup,
	       ProgressInfoUpdateType&	pi_update);

    /// Get an iterator pointing to a specific packet.
    iterator		operator [] (size_t);

    /// Get the number of packets in the trace.
    size_t		size() const;

    /// Get the trace's endianness w.r.t. the host system.
    bool		swapped() const;

    /// Get the trace's time adjustment.
    int32_t		zone() const;

    /// Get the trace's snap length.
    int32_t		snaplen() const;

    /// Get the trace's link type.
    link_type		linktype() const;

    /**
     * Throw an instance of @c std::invalid_argument if the
     * descriptor's link type is different from @c l.
     */
    void		expect(link_type l) const;

    /// Get the trace size in bytes.
    std::streampos	file_size() const;

    /// Get the trace's name.
    const std::string&	file_name() const;

    /// Get a dumper object for a given file.
    pcapxx::dumper	dumper(const std::string& fname) const;

  private:
    template <typename ProgressInfoSetupType,
	      typename ProgressInfoUpdateType>
    void
    setup(const std::string&			filename,
	  bool					build_index,
	  ProgressInfoSetupType&		pi_setup,
	  ProgressInfoUpdateType&		pi_update);

    void		warn(const std::streampos&p, const std::string& msg);
    void		handle_truncation(const std::string& msg);

    template <typename ProgressInfoSetupType,
	      typename ProgressInfoUpdateType>
    void
    setup_marks(std::istream&		f,
		ProgressInfoSetupType&	pi_setup,
		ProgressInfoUpdateType&	pi_update);

    index_type		marks_;

    bool		swap_;
    int32_t		zone_;
    int32_t		snaplen_;
    link_type		type_;
    size_t		pkt_count_;
    std::streampos	file_size_;
    std::string		file_name_;
  };

} // End of namespace pcapxx.

# include "descriptor.hxx"

#endif // ! PCAP_DESCRITPOR_HH_
