/*
 * WiPal - A library and a set of tools to manipulate wireless traces.
 * Copyright (C) 2007, 2008  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */
#ifndef TOOL_OPTIONS_HH_
# define TOOL_OPTIONS_HH_

# include <stdexcept>
# include <string>
# include <map>
# include <ostream>

# include <boost/any.hpp>


namespace tool
{

  namespace opt
  {

    /*---------------------.
    | Forward declarations |
    `---------------------*/

    namespace internals
    {
      struct pair;
      template <class> struct defaults_list;
    }

    struct name;


    /*-------------.
    | Declarations |
    `-------------*/

    /// Exception raised when converting opt::value to invalid types.
    struct invalid_type: std::logic_error
    {
      invalid_type(const std::string& opt_type_name,
		   const std::string& req_type_name);
    };

    /// Option value for opt::list.
    struct value: public boost::any
    {
      value();

      template <class T>
      value(const T& v);

      template <class T>
      T&
      get();

      template <class T>
      const T&
      get() const;

      value&
      operator [] (const name& n);

      const value&
      operator [] (const name& n) const;

      std::ostream&
      print(std::ostream& v) const;
    };

    std::ostream&
    operator << (std::ostream& o, const value& v);

    /// Convert an option value to a specific type.
    template <class T>	T&		get(value& v);

    /// Convert an option value to a specific type, const version.
    template <class T>	const T&	get(const value& v);

    /// Option name for opt::list.
    struct name: public std::string
    {
      name(const std::string& n);
      name(const char* n);

      internals::pair
      operator -= (const value& v) const;
    };

    /// Option related internals.  Not designed for users.  @see opt::list.
    namespace internals
    {

      /// (name, value) option pair for opt::list's constructor.
      struct pair
      {
	pair(const name& n, const value& v);

	defaults_list<pair>
	operator , (const pair& p) const;

	template <class Map>
	void
	populate(Map& m) const;

      private:
	const name&	opt_name;
	const value&	opt_value;
      };

      /// Static list of opt::pair's.
      template <class NextDefaultsList>
      struct defaults_list
      {
	defaults_list(const pair& p, const NextDefaultsList& next);

	defaults_list< defaults_list<NextDefaultsList> >
	operator , (const pair& p) const;

	template <class Map>
	void
	populate(Map& m) const;

      private:
	const pair&		this_pair;
	const NextDefaultsList&	next_defaults;
      };

    } // End of namespace tool::opt::internals.

    /// Exception raised when accessing a non-available option in opt::list.
    struct invalid_option: public std::invalid_argument
    {
      invalid_option(const name& n);
    };

    /**
     * List of options.
     *
     * @fixme FIXME: Document.
     */
    struct list: public std::map<name, value>
    {
      list();

      list(const internals::pair& defaults);

      template <class D>
      list(const internals::defaults_list<D>& defaults);

      value&
      operator [] (const name& n);

      const value&
      operator [] (const name& n) const;

      void
      add(const list& l);

      void
      add(const name& n, const value& v);

      list&
      add(const name& n, const list& l);

      void
      populate(list& other) const;

      std::ostream&
      print(std::ostream& o, unsigned depth = 0) const;
    };

    std::ostream&
    operator << (std::ostream& o, const list& l);

  } // End of namespace tool::opt.

} // End of namespace tool.

# include "options.hxx"

#endif // TOOL_OPTIONS_HH_
