/*
 * WiPal - A library and a set of tools to manipulate wireless traces.
 * Copyright (C) 2007, 2008  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */

#include <utility>
#include <stdexcept>
#include <iostream>
#include <cstdlib>
#include <map>

#include <boost/format.hpp>
#include <boost/foreach.hpp>
#include <boost/preprocessor/tuple/elem.hpp>
#include <boost/preprocessor/seq/for_each.hpp>

#include <wipal/tool/args.hh>
#include <wipal/pcap/list_of_traces.hh>
#include <wipal/wifi/frame/dissector/dissector.hh>
#include <wipal/wifi/frame/dissector/default_hooks.hh>
#include <wipal/wifi/frame/frame.hh>
#include <wipal/wifi/frame/mgt.hh>
#include <wipal/wifi/frame/filter/microseconds_stamper.hh>
#include <wipal/wifi/frame/filter/non_noisy_prism.hh>
#include <wipal/phy/empty.hh>
#include <wipal/phy/prism_header.hh>
#include <wipal/phy/rtap_header.hh>

namespace
{

  struct data_src: public wifi::frame::dissector_default_hooks
  {

    data_src(): check_ (false), src_ (0)
    {
    }

    void
    data_hook(const void* f, size_t)
    {
      // Ignore retransmitted frames.
      check_ = not wifi::frame::control_flag(3, f);
    }

    void
    addr_hook(const void*, size_t, unsigned i, const wifi::addr& a)
    {
      if (check_ and 2 == i)
	src_ = &a;
    }

    const wifi::addr*
    src() const
    {
      return src_;
    }

  private:
    bool		check_;
    const wifi::addr*	src_;

  }; // End of data_src.

  template <class PhyHeader, class T>
  bool
  find_data_dups(T& t, const tool::opt::list& o, boost::format& f)
  {
    typedef typename T::const_iterator			const_iterator;
    typedef typename const_iterator::value_type		value_type;
    typedef wifi::frame::dissector<data_src>		dissector;
    typedef std::map<wifi::addr, value_type>		last_data_map;

    last_data_map		map;
    bool			success (true);
    const const_iterator	e = t.end();

    for (const_iterator i = t.begin(); i != e; ++i)
      {
	const dissector	d = wifi::frame::dissect<PhyHeader, data_src>(*i);

	if (d.src())
	  {
	    const std::pair<typename last_data_map::iterator, bool> r =
	      map.insert(std::make_pair(*d.src(), *i));

	    if (r.second)
	      continue;

	    const value_type&	last = r.first->second;

	    if (PhyHeader::eq_80211(*i, last))
	      {
		success = false;
		if (o["ui"]["summary"].get<bool>())
		  std::cout << (f % *i % last
				  % (i->microseconds() - last.microseconds())
				  % r.first->first);
	      }
	    r.first->second = *i;
	  }
      }
    return success;
  }

} // End of anonymous namespace.

int
main(int argc, char *const argv[])
{
  bool	success (true);

  try
    {
      namespace opt = tool::opt;
      namespace args = tool::args;
      namespace def = tool::args::def;

      opt::list opts ((opt::name ("internals") -= opt::list
		       ((opt::name ("endianness") -= def::endianness(),
			 opt::name ("ignore_noisy_prism") -=
						   def::ignore_noisy_prism())),
		       opt::name ("ui") -= opt::list
		       ((opt::name ("headers") -= def::ui_headers(),
			 opt::name ("summary") -= def::ui_summary()))));

      args::parse(argc, argv, opts, args::arbitrary, args::none);

      boost::format	f ("%10s %10s %20s %20s\n");

      BOOST_FOREACH(pcapxx::list_of_traces<> t,
		    opt::get<args::input>(opts["proc"]["input"]))
	{
	  typedef tool::endian::endianness	endianness;

	  if (opt::get<bool>(opts["ui"]["headers"]))
	    {
	      std::cout << t.file_name() << ":\n"
			<< (f % "Frame" % "Dup. of"
			      % "Time diff. (usec)" % "MAC source addr.");
	      std::cout << (f % "=====" % "======="
			      % "=================" % "================")
			<< std::flush;
	    }

	  if (t.empty())
	    continue;

	  if (opt::get<bool>(opts["internals"]["ignore_noisy_prism"]))
	    t.expect(pcapxx::descriptor<>::PRISM_HEADER);

	  const endianness&	end = opt::get<endianness>(opts["internals"]
							   ["endianness"]);

	  switch (const pcapxx::descriptor<>::link_type lt =
		    t.get(0).linktype())
	    {

#define WP_case(R, Unused, LT)						\
	    case pcapxx::descriptor<>::BOOST_PP_TUPLE_ELEM(2, 0, LT):	\
	      {								\
		typedef wifi::frame::filter::microseconds_stamper<	\
			  pcapxx::list_of_traces<>::const_iterator,	\
			  BOOST_PP_TUPLE_ELEM(2, 1, LT)>		\
		  ms_stamper_t;						\
									\
		ms_stamper_t	mss (t.begin(), t.end(), end);		\
									\
		if (opt::get<bool>(opts["internals"]			\
				   ["ignore_noisy_prism"]))		\
		  {							\
		    typedef ms_stamper_t::const_iterator		it_t; \
		    typedef wifi::frame::filter::non_noisy_prism<it_t>	nnp_t; \
									\
		    nnp_t	n (mss.begin(), mss.end());		\
									\
		    success = find_data_dups<BOOST_PP_TUPLE_ELEM(2, 1, LT)> \
		      (n, opts, f) and success;			\
		  }							\
		else							\
		  success = find_data_dups<BOOST_PP_TUPLE_ELEM(2, 1, LT)> \
		    (mss, opts, f) and success;			\
	      }								\
	    break;

	      BOOST_PP_SEQ_FOR_EACH(WP_case, ~, WP_LINKTYPES)

	    default:
	      success = false;
	      args::unsupported_linktype(lt);
	      break;
	    }
	}
    }
  catch (std::exception& e)
    {
      success = false;
      std::cerr << "ERROR: " << e.what() << '.' << std::endl;
    }

  return success ? EXIT_SUCCESS : EXIT_FAILURE;
}
